#include "defs.h"
#include "font.h"
#include "screen.h"
#include "snellen.h"
#include <math.h>
#include <ctype.h>

	/*
	** X things
	*/
Window		topwin;
static Atom     wm_delete_window;
static String	fallback_resources[] = {
    "*input: True",
    "*allowShellResize: True",
    "*Dialog*value: ",
    "*Dialog.label: Subject's name?",
    "*Dialog.width: 400",
    "*Dialog*value.translations: #override \\n <Key>Return: Nop()",
    "*Dialog*label.resizable: TRUE",
    "*SimpleMenu.label: Hello there",
    NULL,
};
GC		makegc();
void		DoOneSize(int f);
void		Nop();
float		randfrac();
void		initpositions();
char		charbuf[BUFSIZ];
char		getreplay();
int 		direction, font_ascent, font_descent;
XCharStruct	xcs;
int		height;
int		target_height;
int		width;
int		xpos;
int		ypos;
int		xoffset[9] = { 0, 1,  0, -1, 0, 1,  0, -1, 0};
int		yoffset[9] = { 0, 0, -1,  0, 1, 0, -1,  0, 1};
float		frac;
static XtActionsRec actions[] =
{
	{"Start",	DoTest	},
	{"Nop",		Nop	},
};
XtEventHandler          ExposeCB();

struct _app_resources
{
	Boolean	enableBackups;
	char	*backupNamePrefix;
	char	*backupNameSuffix;
} app_resources;
#define         Offset(field) XtOffsetOf(struct _app_resources, field)
static XtResource	resources[] =
{
	{"enableBackups", "EnableBackups", XtRBoolean, sizeof(Boolean),
		Offset(enableBackups), XtRImmediate, FALSE},
};
char			*version = "xacuity version 0.30 23 November 2005 - Kettering";
int			testing;
int			limit = 9;
int			n_trials = 10;

int
main(int argc, char **argv)
{
	XtAppContext		appcon;
	XTextProperty		icon_prop;
	char			*icon_name[2];
	time_t			now;

	if((progname = strrchr(*argv,'/')) != NULL)
		progname++;
	else
		progname = *argv;

	icon_name[0] = progname;
	icon_name[1] = 0;
	(void) time(&now);
	(void) srand(now);
	start_tm = localtime(&now);
	centre_frequency = 1;
	duration = 100;
	person = "unknown";
	strcpy(eyename, "unknown");
	
	while(--argc)
	{
		if(**++argv == '-')
		    switch(*++*argv)
		{
			case 'D':
				debug++;
				continue;
			case 'a':
				if(*++*argv == '\0')
				{
					argc--;
					argv++;
				}
				diameter = atof(*argv);
				warn("diameter set from argument now %f",
					diameter);
				continue;
			case 'c':
				if(*++*argv == '\0')
				{
					argc--;
					argv++;
				}
				centre_frequency = atoi(*argv);
				if(centre_frequency < 1)
					ferr("centre frequency must be 1 or more");
				continue;
			case 'd':
				if(*++*argv == '\0')
				{
					if(argc <= 1 || *argv[1] == '-')
					{
						ferr("-d requires duration");
					}
					argc--;
					argv++;
				}
				duration = atoi(*argv);
				continue;
			case 'i':
				argc--;
				argv++;
				if((replayfp = fopen(*argv,"r")) == NULL)
					ferr("cannot open %s for reading",
						*argv);
				replaying = 1;
				ReadHeaders(replayfp, *argv);
				fprintf(stderr,"in %s eye was %s\n",
					*argv, eyename);
				continue;
			case 'l':
				argc--;
				argv++;
				limit = atoi(*argv);
				if
				(
					limit != 1
					&&
					limit != 5
					&&
					limit != 9
				)
					ferr("limit must be 1, 5 or 9");
				continue;
			case 'n':
				if(*++*argv == '\0')
				{
					argc--;
					argv++;
				}
				n_trials = atoi(*argv);
				if(n_trials <= 0)
					ferr("number of trials must be > 0");
				continue;
			case 'p':	
				if(*++*argv == '\0')
				{
					argc--;
					argv++;
				}
				start_point_size = atof(*argv);
				continue;
			case 'r':
				reverse = 1;
				continue;
			case 't':
				testing = 1;
				continue;
			case 'v':
				warn("%s", version);
				exit(0);
			case '?':
				warn("usage: -p starting_point_size");
				exit(0);
			default:
				ferr("unknown flag: %c", **argv);
		}
	}

	/*
	** initialise X
	*/
	top = XtAppInitialize
	(
		&appcon,
		"Xacuity",
		NULL,
		0,
		&argc,
		argv,
		fallback_resources,
		NULL,
		0
	);
	XtAppAddActions(appcon, actions, XtNumber(actions));
	XtGetApplicationResources
	(
		top,
		(XtPointer) &app_resources,
		resources,
		XtNumber(resources),
		NULL,
		0
	);
	display = XtDisplay(top);

	makeButtonsAndBoxes(top);

	XtRealizeWidget(top);

	XDefineCursor
	(
		XtDisplay(top),
		(topwin = XtWindow(top)),
		XCreateFontCursor(XtDisplay(top), XC_left_ptr)
	);
	XStoreName(XtDisplay(top), topwin, "X Acuity");
	XStringListToTextProperty(icon_name, 1, &icon_prop);
	XSetWMIconName(XtDisplay(top), topwin, &icon_prop);
	(void) XSetWMProtocols
	(
		XtDisplay(top),
		XtWindow(top),
		&wm_delete_window,
		1
	);

	graphics_window = XtWindow(graphics_widget);
	graphics_gc = makegc
	(
		PAINT,
		MYBLACK,
		MYWHITE
	);
	clobber_gc = makegc
	(
		ERASE,
		MYBLACK,
		MYWHITE
	);

	if(!replaying)
	{
		InfoPrintf("Complete the dialog box");
		PersonPopup(graphics_widget);
	}
	fprintf(stderr,"adding ExposeCB\n");
	XtAddEventHandler(graphics_widget, ExposureMask, False,
		(XtEventHandler)ExposeCB, NULL);
	XtAppMainLoop(appcon);
	return(0);
}

#include <stdarg.h>
void ferr(char *fmt, ...)
{
	va_list	vp;

	fprintf(stderr,"%s: ", progname);
	va_start(vp, fmt);
	vfprintf(stderr,fmt, vp);
	va_end(vp);
	fprintf(stderr,"\n");
	exit(1);
}

void warn(char *fmt, ...)
{
	va_list	vp;

	va_start(vp, fmt);
	vfprintf(stderr,fmt, vp);
	va_end(vp);
	fprintf(stderr,"\n");
}

char getreplay()
{
	char	linbuf[BUFSIZ];
	int	count;

	if(fgets(linbuf, BUFSIZ, replayfp) == NULL)
		return('\033');
	fprintf(stderr,"read <%s>\n", linbuf);
	count = sscanf(linbuf, "%d %d %c %c", 
		&replaypos, &replayfont, &replaychar, &replayans);
	fprintf(stderr,"got %d: %d %d %c %c\n",
		count,  replaypos, replayfont, replaychar, replayans);
	if(count != 4)
		return('\033');
	else
		return(' ');
}

void DoTest()
{
	char			kbuf;
	char			snellen();
	int			npos;		/* 0-4 */
	int			nposition;	/* 0-8 */
	int			i;
	int			randpos();
	struct position		*pp;
	int			np;
	int			finished;

	initfonts();
	initpositions();
	OpenTmpFile();

	charbuf[0] = 'E';
	charbuf[1] = '\0';

	if(StartedButNotSaved)
	{
		ErrorPopup(graphics_widget, "Last run not saved");
		StartedButNotSaved = 0;
		return;
	}

	DrawTarget();

	if(!replaying)
		InfoPrintf("Press the spacebar when ready");
	else
		InfoPrintf("Replaying data");
	XFlush(display);
	while(1) if
	(
		( replaying && ((kbuf = getreplay()) != (char)NULL) )
		||
		(
		! replaying
		&&
		XCheckWindowEvent(display, graphics_window,
			ExposureMask| KeyPressMask, &event) == True
		&&
		XLookupString((XKeyEvent *)&event, &kbuf, 1, NULL, NULL) == 1
		)
	)
	switch(kbuf)
	{
		case '\033':
			if(npos)
				StartedButNotSaved++;
			InfoPrintf("Use File menu to save/quit");
			return;
		case '\06':
		case '\f':
			DrawTarget();
			continue;
		case ' ':
			if(! replaying)
				charbuf[0] = snellen();
			else
				charbuf[0] = replaychar;
another_position:
			if(! replaying)
			{
				nposition = npos = randpos();
				frac = randfrac();
				if(nposition)
					nposition = npos + 4*(frac == 1.);
			}
			else
				nposition = replaypos;
			pp = &position[nposition];
			if(((pp->flag & FLAG_BLIND) || pp->done) && ! replaying)
				goto another_position;
			if(! replaying)
				nfont = pp->nfont;
			else
				nfont = replayfont;
			setfonts(nfont);
			XTextExtents(fontinfo[nfont], charbuf, 1,
				&direction, &font_ascent,
				&font_descent, &xcs);
			width = xcs.width;
			height = font_ascent;
			xpos = pp->x - width/2;
			ypos = pp->y + height/2;
			XDrawString(display, graphics_window, graphics_gc,
				 xpos, ypos, charbuf, 1);
			XFlush(display);
			usleep((long)1000*duration);
			XDrawString(display, graphics_window, clobber_gc,
				 xpos, ypos, charbuf, 1);
			XFlush(display);
			InfoPrintf("Type the character you saw; spacebar if unsure");
			if(!replaying)
			while(XCheckWindowEvent(display, graphics_window,
				ExposureMask|
				KeyPressMask, &event) != True);
			if(
				(
				! replaying
				&&
				XLookupString((XKeyEvent *)&event, 
				&kbuf, 1, NULL, NULL) == 1
				)
				||
				(
				replaying
				&&
				(kbuf = replayans) != (char)NULL
				)
			)
			{
				if(debug)
				fprintf(stderr,"Got %c\n", kbuf);
				if(kbuf>= 'a' && kbuf <= 'z')
					kbuf = kbuf - 'a' + 'A';
				if(kbuf == ' ')
					kbuf = '-';
				if(!isprint(kbuf))
					kbuf = '!';

				if(tmpfp)
				{
					fprintf
					(
						tmpfp,
						"%d %d %c %c\n",
						nposition,
						nfont,
						charbuf[0],
						kbuf
					);
					fflush(tmpfp);
				}
				pp->testnum[nfont][pp->ntests[nfont]] = testnum++;
				pp->seqnum[nfont][pp->ntests[nfont]] = seqnum[nposition]++;
				pp->testans[nfont][pp->ntests[nfont]] = kbuf;
				pp->testchar[nfont][pp->ntests[nfont]] = charbuf[0];
				pp->testsize[nfont][pp->ntests[nfont]] = point_size[nfont];
				if(pp->ntests[nfont]++ == MAXTESTS)
				{
					fprintf(stderr,"too many tests - forced exit");
					goto problem_exit;
				}

				for(i = BLIND_CRITERION - 1; i> 0; i--)
					pp->lastresult[i] = pp->lastresult[i-1];
				/*
				** special things for correct result
				*/
				if(kbuf == charbuf[0])
				{
					pp->lastresult[0] = 1;
					pp->ncorrect[nfont]++;
					/* special case for font above first error */
					/* added in 0.25 */
					if( pp->anyerror == 1)
					{
						if(pp->ntests[nfont] >= n_trials)
							pp->anyerror = 2;
						else
						{
						InfoPrintf("Press the spacebar when ready or Esc to stop");
						XFlush(display);
						continue;
						}
					}
					if
					(
						/* first test correct */
						pp->ntests[nfont] == 1
						&&
						nfont +1 <maxfontsizes /* 0.20 added +1 */
					)
					{
						pp->nfont++;
						if(debug)
							fprintf(stderr, "pp->nfont now = %d\n", pp->nfont);
						/* 0.25 added the InfoPrintf here */
						InfoPrintf("Press the spacebar when ready or Esc to stop");
						XFlush(display);
						continue; /* 0.19 added */
					}
				}
				else /* subject got it wrong */
				{
					pp->lastresult[0] = -1;
					/* special case - still first font */
					if(pp->nfont == 0)
					{
						int failure = 1;
						for(i = 1; i < BLIND_CRITERION; i++)
							if(pp->lastresult[i] != -1)
								failure = 0;
						if (failure)
						{
							pp->flag |= FLAG_BLIND;
							pp->flag |= FLAG_FINISHED;
							pp->done = 1;
							fprintf(stderr,"position %d flagged blind\n", nposition);
						}
					}
					else if(pp->anyerror == 0)
					{
						fprintf(stderr,"anyerror=0 code\n");
						pp->anyerror = 1;
						pp->nfont--;
						InfoPrintf("Press the spacebar when ready or Esc to stop");
						XFlush(display);
						continue;
					}
				}
				/* we are not finished
					- do we change the font size? */
				if(!pp->done && pp->ntests[nfont] >= n_trials)
				{
					/* > 75% */
					if(pp->ncorrect[nfont] >=
						((int)(0.75*n_trials +0.5)))
					{
						if((pp->flag & FLAG_BELOW_DONE)
							|| nfont == maxfontsizes - 1)	/* 0.19 added " - 1" */
						{
							pp->flag |= FLAG_FINISHED;
							pp->done = 1;
						}
						else
						{
							pp->nfont++;
							if(debug)
								fprintf(stderr,"nfont now = %d\n", pp->nfont);

						}
					}
					else /* < 75% */
					{
						pp->flag |= FLAG_BELOW_DONE;
						if(pp->nfont && pp->ntests[nfont-1] < n_trials)
						{
							if(--pp->nfont < 0)
							{
								pp->flag |= FLAG_FINISHED;
								pp->done = 1;
							}
							if(pp->nfont == 0)
								for(i = 0; i < BLIND_CRITERION; i++)
									pp->lastresult[i] = 0;
						}
						else
							pp->done = 1;
					}
				}
				for(np = 0, finished = 0; np < limit; np++)
					finished += position[np].done;
				if(finished == limit)
				{
problem_exit:
					InfoPrintf("Test finished.  SAVE DATA!!!");
					return;
				}
			}
			InfoPrintf("Press the spacebar when ready or Esc to stop");
			XFlush(display);
			break;

			
		default:
			fprintf(stderr,"%c (0%o) ignored\n", kbuf, kbuf);
			break;
	}
}

void
DoQuit()
{
	if(StartedButNotSaved)
	{
		ErrorPopup(graphics_widget, "Last run not saved");
		StartedButNotSaved = 0;
		return;
	}
	exit(0);
}

void
DoPattern ()
{
	int	i;

	if(! fontsready)
	{
		initfonts();
		initpositions();
	}

	for(i = 0; i< maxfontsizes; i++)
	{
		DoOneSize(i);
		sleep(1);
	}
	DoOneSize(0);
}

void DoOneSize(int f)
{
	int	i;
	int	npos;

	DrawTarget();
	setfonts(f);
	nfont = f;
	XTextExtents(fontinfo[nfont], charbuf, 1,
		&direction, &font_ascent,
		&font_descent, &xcs);
	width = xcs.width;
	height = font_ascent;

	for(i=0; i<=8; i++)
	{
		charbuf[0] = '0' + i;
		npos = i;
		frac = 1.0;
		if(npos > 4)
		{
			npos -= 4;
			frac = FRACTION;
		}
		xpos = position[i].x - width/2;
		ypos = position[i].y + height/2;
		XDrawString(display, graphics_window, graphics_gc, xpos, ypos, charbuf, 1);
		XFlush(display);
	}
}

void
InfoPrintf(char *str)
{
	XawTextBlock		text;
	static int		last_length = 0;
	char			lbuf[BUFSIZ];

	sprintf(lbuf, " %s", str);

	text.length = strlen(lbuf);
	text.ptr = lbuf;
	text.firstPos = 0;
	text.format = FMT8BIT;
	if(text.length == 0)
		return;
	if(*(lbuf + text.length-1) == '\n')
	{
		*(lbuf + text.length-1) = '\0';
		text.length--;
	}
	if(info_text == (Widget)0)
	{
		fprintf(stderr,"%s has no message widget to use\n", progname);
		fprintf(stderr,"%s", lbuf);
		return;
	}
	XawTextReplace( info_text, 0, last_length, &text);
	last_length = text.length;
}

int randpos()
{
	int	r;
	int	p;

	if(limit == 1)
		return(0);
	r = rand();
	p = (int)(( (4.+centre_frequency) *r)/RAND_MAX);
	if(p > 4)
		p = 0;
	return(p);
}

float randfrac()
{
	int	r;
	int	p;
	float	rings = 2.0;
	
	if(limit == 5)
		rings = 1.0;

	r = rand();
	p = (int)(( (rings) *r)/RAND_MAX);
	if(p < 1)
		return(FRACTION);
	return(1.);
}

char	snellen()
{
	int	r;
	int	integer;
	char	c;

	r = rand();
	c = snellen_chars[integer = (int)((((float)N_SNELLEN)*r)/RAND_MAX)];
	return(c);
}

void DrawTarget()
{
	int	line_style = 0; /* solid */
	int	cap_style = CapButt;
	int	join_style = JoinRound;
	int	line_width = 1;
	int	sign1;
	int	sign2;

	if(debug)
		fprintf(stderr,"DrawTarget called\n");
	/* if we are called by an expose event before initialisation
		do nothing */
	if(!fontsready)
		return;

	XClearWindow(display,graphics_window);
	XSelectInput
	(
		display,
		graphics_window,
		KeyPressMask|ButtonPressMask
		|ExposureMask
	);
	XFlush(display);
	if(!target_height)
	{
		XTextExtents(fontinfo[nfont], charbuf, 1,
			&direction, &font_ascent, &font_descent, &xcs);
		target_height = font_ascent;
	}
	XSetLineAttributes(display, graphics_gc, line_width, line_style,
		cap_style, join_style);
	for(sign1 = -1; sign1 <= 1; sign1+=2)
		for(sign2 = -1; sign2 <= 1; sign2 +=2)
	{
		XDrawLine(display, graphics_window,
			graphics_gc,
			WINMID+ sign1*target_height/2,
			WINMID+ sign2*target_height/2,
			WINMID+ sign1*target_height/2 + sign1*target_height,
			WINMID+ sign2*target_height/2 + sign2*target_height);
	}
	line_width = 4;
	XSetLineAttributes(display, graphics_gc, line_width, line_style,
		cap_style, join_style);
	
	for(sign1 = -1; sign1 <= 1; sign1+=2)
		for(sign2 = -1; sign2 <= 1; sign2 +=2)
	{
		XDrawLine(display, graphics_window,
			graphics_gc,
			WINMID+ sign1*target_height,
			WINMID+ sign2*target_height,
			WINMID+ sign1*target_height + sign1*target_height*2,
			WINMID+ sign2*target_height + sign2*target_height*2);
	}
}

void initpositions()
{
	int i;

	position[0].x = WINMID;
	position[0].y = WINMID;

	i=0;
	if(debug)
		fprintf(stderr,"%d: %d %d\n", i, position[i].x, position[i].y);

	for(i = 1; i < limit; i++)
	{
		position[i].x = WINMID +
			(diameter/2.0)*xoffset[i]*(i > 4? FRACTION : 1.);
		position[i].y = WINMID +
			(diameter/2.0)*yoffset[i]*(i > 4? FRACTION : 1.);
		if(debug)
			fprintf(stderr,"%d: %d %d\n", i,
				position[i].x, position[i].y);
	}
	if (testing)
	for(i = 1; i < limit; i++)
	{
		position[i].done = 1;
		position[i].flag = FLAG_FINISHED;
	}
}
